"""
Providers module.
"""

from .exceptions import FactoryError, UndefinedDependencyError


class Provider(object):
    """
    Dependency provider.
    """

    def __init__(self, key):
        """
        Constructor.

        :param str key:
        :return:
        """
        self.key = key

    def __call__(self, *args, **kwargs):
        """
        Provides instance.

        :return:
        """
        raise NotImplementedError()


class Factory(Provider):
    """
    Factory provider produces new instance on every call.
    """

    def __init__(self, cls, dependencies=None, key=None, producer=None):
        """
        Constructor.

        :param type cls:
        :param list[pybinder.requires.Requires] dependencies:
        :param str key:
        :return:
        """
        self.cls = cls
        if not dependencies:
            dependencies = list()
        self.dependencies = dependencies
        if not producer:
            producer = self._default_producer
        self.producer = producer
        super(Factory, self).__init__(key)

    def __call__(self, *context_args, **context_kwargs):
        """
        Provides instance.
        """
        injected_kwargs = dict((
            (requires.with_name, requires.provide())
            for requires in self.dependencies
        ))
        kwargs = dict()
        kwargs.update(injected_kwargs)
        kwargs.update(context_kwargs)
        try:
            return self.producer(self.cls, *context_args, **kwargs)
        except TypeError as exception:
            raise FactoryError(exception.args[0],  context_args,
                               context_kwargs, injected_kwargs)

    def assemble(self, dependencies):
        """
        Assembles and gets ready to work.

        :param dependencies:
        :return:
        """
        def inject_provider(requirement):
            if requirement.provider is not None:
                return requirement
            namespace, key = requirement.from_namespace, requirement.key
            try:
                requirement.provider = dependencies[namespace][key]
            except KeyError:
                raise UndefinedDependencyError(
                    'Required dependency "{0}" '.format(key) +
                    'could not be found in namespace "{0}"'.format(namespace)
                )
            return requirement

        self.dependencies = [inject_provider(requirement)
                             for requirement in self.dependencies]

    @property
    def dependent_providers(self):
        """
        Returns dependent providers.

        :return:
        """
        return [requires.provider for requires in self.dependencies]

    @staticmethod
    def _default_producer(cls, *args, **kwargs):
        """
        Default producer.

        :param cls:
        :param args:
        :param kwargs:
        :return:
        """
        return cls(*args, **kwargs)


class Singleton(Factory):
    """
    Singleton provider produces new instance once and returns it on every call.
    """

    def __init__(self, cls, dependencies=None, key=None):
        """
        Constructor.

        :return:
        """
        self.instance = None
        super(Singleton, self).__init__(cls, dependencies, key)

    def __call__(self, *args, **kwargs):
        """
        Provides instance.
        """
        if self.instance is None:
            self.instance = super(Singleton, self).__call__(*args, **kwargs)
        return self.instance


class _ObjectProvider(Provider):
    """
    Object provider memorizes something once and returns it on every call.
    """

    def __init__(self, what, key=None):
        """
        Constructor.

        :param what:
        :param str key:
        :return:
        """
        self.what = what
        super(_ObjectProvider, self).__init__(key)

    def __call__(self):
        """
        Provides instance.
        """
        return self.what


class Value(_ObjectProvider):
    """
    Value provider memorizes value once and returns it on every call.
    """


class Object(_ObjectProvider):
    """
    Object provider memorizes object once and returns it on every call.
    """


class Class(_ObjectProvider):
    """
    Class provider memorizes class once and returns it on every call.
    """


class Function(_ObjectProvider):
    """
    Function provider memorizes function once and returns it on every call.
    """
